#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pyrtz']

package_data = \
{'': ['*']}

install_requires = \
['numpy >=1.18.5',
 'pandas >=1.0.5',
 'scipy >=1.4.1',
 'plotly >=5.1.0',
 'dash >=1.21.0',
 'dash_core_components >=1.17.1',
 'dash_html_components >=1.1.4',
 'PyPDF2 >=1.26.0',
 'igor >=0.3',
 'kaleido >=0.2.1']

setup(name='pyrtz',
      version='0.1.3',
      description='Force spectroscopy in Python',
      author='Nicholas Stone',
      author_email='stone.nicholas@gmail.com',
      url='https://github.com/nstone8/pyrtz',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.8',
     )
